#!/usr/bin/perl
#Copyright (c) 2019 Alexandr A. Teyatnikov, http://alter.org.ua

$test=0;
$migrate=0;
$cp="CP1251";
$import = 0;
$root_key = "./.far2l.REG";

print "\n";

NEXT_ARG:
for($i=0; !(($v = $ARGV[$i]."") eq ""); $i++) {
  print "v=$v\n";
  if($v eq "-h" || $v eq "--help") {
    print "Import saved Far settings from Windows .REG files\n".
          "Usage:\n".
          "  Far2lRestore.pl [-h|--help]\n".
          "  Far2lRestore.pl [<options>] FarSave1.reg\n".
          "Options:\n".
          "  -m|--migrate    copy setting from Far 1.x hive (HCU\\Software\\Far).\n".
          "                    default HCU\\Software\\Far2\n".
          "  -t|--test       just display imported information, but do nothing\n".
          "  -c|--cp <src encoding>\n".
          "                  use specified encoding for REGEDIT4 files\n".
          "                    default CP1251\n".
          "  -i|--import     store parsed info under ~/.far2l/REG\n".
          "                    default .far2l.REG in current directory\n".
          "";
    exit;
  }
  if($v eq "-m" || $v eq "--migrate") {
    $migrate=1;
    next NEXT_ARG;
  }
  if($v eq "-t" || $v eq "--test") {
    $test=1;
    next NEXT_ARG;
  }
  if($v eq "-i" || $v eq "--import") {
    $import=1;
    $root_key = $ENV{'HOME'}."/.far2l/REG";
    next NEXT_ARG;
  }
  if($v eq "-c" || $v eq "--cp") {
    $i++;
    $v=$ARGV[$i];
    if($v eq "") {
      print "Missing encoding name\n";
      exit;
    }
    $cp=$v;
    next NEXT_ARG;
  }
  $in_name = $v;
}

if(! -d $root_key) {
  mkdir($root_key);
}

if(! -s $in_name) {
  print "Cannot open source file $in_name\n";
  exit;
}

open(in_file, "<".$in_name) or die ;
if (<in_file>) {
  $s = $_;
  $s =~ s/\0//g;
  if($s =~ /Windows Registry Editor Version 5.00/) {
    $cp = "UCS-2LE";
  }
}
close(in_file);

$cmd = "iconv -f $cp -t utf8 $in_name > $in_name.tmp";
print $cmd."\n";
system($cmd);
print "Open $in_name.tmp\n";
open(in_file, "<".$in_name.".tmp") or die ;

$cur_key = "";
$in_hex = 0;
$cur_name = "";
$cur_v = "";

NEXT_LINE:
while (<in_file>) {
  $s = $_;
  $s =~ s/\r//g;

  if($in_hex) {
    if(handle_hex($s)) {
      next NEXT_LINE;
    }
  }
  if($s =~ /^\[(HKEY_[a-zA-Z\-_\\\d]+.*)\]$/) {
    $cur_key = $1;
    #print $cur_key."\n";
    if($migrate) {
      $cur_key =~ s/\\Software\\Far\\/\\Software\\Far2\\/i;
      $cur_key =~ s/\\Software\\Far$/\\Software\\Far2/i;
    }
    $cur_key =~ s/\\/\/k-/g;
    $cur_key =~ s/HKEY_LOCAL_MACHINE/HKLM/;
    $cur_key =~ s/HKEY_CURRENT_USER/HKU\/c/;
    if( -d $root_key."/".$cur_key ) {
      next NEXT_LINE;
    }
    $tmp_key = "";
    print $cur_key.":\n";
    @keys = split(/\//, $cur_key);
  NEXT_KEY: foreach $str (@keys) {
      $tmp_key .= "/".$str;
      if( ! -d $root_key.$tmp_key ) {
        if($test) {
          print $tmp_key."\n";
        } else {
          mkdir($root_key.$tmp_key);
        }
      }
    }
    next NEXT_LINE;
  }
  if($cur_key eq "") {
    next NEXT_LINE;
  }

  if($s =~ /^\"([a-zA-Z\-_\\\d]+)\"=(.*)/) {
    $cur_name = $1;
    $v = $2;
    $cur_v = "";
    print $cur_name.":\n";
    if($v =~ /^hex:(.*)/) {
      $s1 = $1;
      handle_hex($s1);
      next NEXT_LINE;
    }
    if($v =~ /^dword:(.*)/) {
      $v = $1;
      $v =~ s/^0*//g;
      if($v eq "") {
        $v = 0;
      }
      flush_v($cur_key, $cur_name, "DWORD", $v);
      next NEXT_LINE;
    }
    if($v =~ /^\"(.*)\"$/) {
      $v = $1;
      $v =~ s/\\\"/\"/g;
      $v .= "\\0";
      flush_v($cur_key, $cur_name, "SZ", $v);
      next NEXT_LINE;
    }
    print "Unknown value type:'n". "'".$cur_name."/".$v."'\n";
  }
}

close(in_file);
unlink("$in_name.tmp");

sub handle_hex
{
  my $s = $_[0];
  if($s =~ /\s*([[a-zA-Z,\d]+)(\s|,\\|$)/) {
    my $v = $1;
    my $eol = $2;
    $v =~ s/,/ /g;
    $cur_v .= $v."\n";
    if($eol eq ",\\") {
      $in_hex = 1;
      return 1;
    }
    flush_v($cur_key, $cur_name, "BINARY", $cur_v);
    $cur_v = "";
    $cur_key = "";
    $in_hex = 0;
    return 1;
  } else {
    $in_hex = 0;
  }
  return 0;
}

sub flush_v
{
  my $key = $_[0];
  my $name = $_[1];
  my $type = $_[2];
  my $v = $_[3];

  my $i=0;
  my @a = split(/[\n\r\s]+/, $v);
  $v = "";
  foreach $octet (@a) {
    if($i % 16) {
      $v .= " ";
    }
    $v .= $octet;
    $i++;
    if(!($i % 16)) {
      $v .= "\n";
    }
  }
  if($i % 16) {
    $v .= "\n";
  }

  my $d = $name."\n".
          $type."\n".
          $v;

  my $fn = $root_key."/".$key."/v-".$name;
  if($test) {
    print $fn.":\n";
    print $d;
  } else {
    open(out_file, ">".$fn);
    print out_file $d;
    close(out_file);
  }
}
